/*
   3rd Person Camera Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to build off of the previous camera demos and
   it will add the use of a basic third person view.  To do this we will overload the
   RotateCamera function to allow the camera to rotate around a point.  This point
   will be the view.  We will use the view to translate our object so we can have the
   camera follow it.  This camera is a simple 3rd person camera and would require more
   work to get it at a point where it could be used with all types of third person games
   but this is a start.  It also uses quaternion rotations.
*/


// Demo: 3rd Person Camera.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.

#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include"CTGALoader.h"
#include"CCamera.h"


// Function prototypes...
bool InitializeGL();       // Initialize OpenGL.
void RenderScene();        // Renders each frame.
void GetInput();           // Check the input for each frame.
void Shutdown();           // Release all resources.


// The first is a global HDC that will be used in the RenderScene() function to swap buffers.
HDC g_HDC;

CTGALoader Terrain;        // Terrain texture image.
CTGALoader CubeImage;      // Cube's texture image.

CCamera Camera;            // Camera object used to move around the 3D scene.


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.


   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;

         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;

         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // Calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;

         // This will be used to escape from the app.
         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;
               }
            break;

         default: // Always have a default just in case.
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "3rd Person Camera Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640,
                        480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   void Shutdown();

   return (int)msg.wParam;             // End the application.
}


bool InitializeGL()
{
   glClearColor(0.5f, 0.5f, 1.0f, 1.0f);  // Clear the screen to the specified color.

   glShadeModel(GL_SMOOTH);               // Smooth shading in our scenes.
   glEnable(GL_DEPTH_TEST);               // Enable desth testing for hidden surface removal.
   glEnable(GL_TEXTURE_2D);               // Enable texture mapping.

   // Load the terrain texture.   
   if(!Terrain.LoadTGA("ground.tga")) return false;
   glGenTextures(1, &Terrain.ID);
   
   glBindTexture(GL_TEXTURE_2D, Terrain.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   int type = Terrain.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, Terrain.imageWidth, Terrain.imageHeight,
                     type, GL_UNSIGNED_BYTE, Terrain.image);

   // Load the texture that will be displayed on the screen.
   if(!CubeImage.LoadTGA("Cube.tga")) return false;
   glGenTextures(1, &CubeImage.ID);
   
   glBindTexture(GL_TEXTURE_2D, CubeImage.ID);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

   type = CubeImage.type == 3 ? GL_RGB : GL_RGBA;
   gluBuild2DMipmaps(GL_TEXTURE_2D, type, CubeImage.imageWidth, CubeImage.imageHeight,
                     type, GL_UNSIGNED_BYTE, CubeImage.image);

   // Here we set the camera's default position.
   Camera.SetCamera(0.0f, 0.2f, -2.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   // The get input function will check for input for each frame.
   GetInput();

   // gluLookAt() will position the camera.
   gluLookAt(Camera.Pos.x, Camera.Pos.y, Camera.Pos.z,
             Camera.View.x, Camera.View.y, Camera.View.z,
             Camera.Up.x, Camera.Up.y, Camera.Up.z);


   // Draw the cube.
   glPushMatrix();

      // Position our object.  Our 3rd person camera will be following this.
      glTranslatef(Camera.View.x, 0, Camera.View.z);
   
      // Bind our cube texture to the cube.
      glBindTexture(GL_TEXTURE_2D, CubeImage.ID);

      // Front face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(0.3f, 0.0f, 0.0f);
         glTexCoord2f(1.0, 0.0); glVertex3f(-0.2f, 0.0f, 0.0f);
         glTexCoord2f(1.0, 1.0); glVertex3f(-0.2f, 0.5f, 0.0f);
         glTexCoord2f(0.0, 1.0); glVertex3f(0.3f, 0.5f, 0.0f);
      glEnd();

      // Back face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(0.3f, 0.0f, -0.5f);
         glTexCoord2f(1.0, 0.0); glVertex3f(-0.2f, 0.0f, -0.5f);
         glTexCoord2f(1.0, 1.0); glVertex3f(-0.2f, 0.5f, -0.5f);
         glTexCoord2f(0.0, 1.0); glVertex3f(0.3f, 0.5f, -0.5f);
      glEnd();

      // Left face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(-0.2f, 0.0f, -0.5f);
         glTexCoord2f(0.0, 1.0); glVertex3f(-0.2f, 0.5f, -0.5f);
         glTexCoord2f(1.0, 1.0); glVertex3f(-0.2f, 0.5f, 0.0f);
         glTexCoord2f(1.0, 0.0); glVertex3f(-0.2f, 0.0f, 0.0f);
      glEnd();

      // Right face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(0.3f, 0.0f, -0.5f);
         glTexCoord2f(0.0, 1.0); glVertex3f(0.3f, 0.5f, -0.5f);
         glTexCoord2f(1.0, 1.0); glVertex3f(0.3f, 0.5f, 0.0f);
         glTexCoord2f(1.0, 0.0); glVertex3f(0.3f, 0.0f, 0.0f);
      glEnd();

      // Top face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(-0.2f, 0.5f, 0.0f);
         glTexCoord2f(0.0, 1.0); glVertex3f(-0.2f, 0.5f, -0.5f);
         glTexCoord2f(1.0, 1.0); glVertex3f(0.3f, 0.5f, -0.5f);
         glTexCoord2f(1.0, 0.0); glVertex3f(0.3f, 0.5f, 0.0f);
      glEnd();

      // Bottom face.
      glBegin(GL_QUADS);
         glTexCoord2f(0.0, 0.0); glVertex3f(-0.2f, 0.0f, 0.0f);
         glTexCoord2f(0.0, 1.0); glVertex3f(-0.2f, 0.0f, -0.5f);
         glTexCoord2f(1.0, 1.0); glVertex3f(0.3f, 0.0f, -0.5f);
         glTexCoord2f(1.0, 0.0); glVertex3f(0.3f, 0.0f, 0.0f);
      glEnd();

   glPopMatrix();

   // These two will be used with drawing the terrain.
   int x, z = 0;

   // Draw the ground.
   glBindTexture(GL_TEXTURE_2D, Terrain.ID);
   for(int i = 0; i < 10; i++)
      {
         for(int j = 0; j < 10; j++)
            {
               glBegin(GL_QUADS);
                  glTexCoord2f(0.0, 0.0); glVertex3f(15.0f + x, -0.01f, 15.0f + z);
                  glTexCoord2f(1.0, 0.0); glVertex3f(10.0f + x, -0.01f, 15.0f + z);
                  glTexCoord2f(1.0, 1.0); glVertex3f(10.0f + x, -0.01f, 10.0f + z);
                  glTexCoord2f(0.0, 1.0); glVertex3f(15.0f + x, -0.01f, 10.0f + z);
               glEnd();

               x += -5;
            }
         x = 0;
         z += -5;
      }
 
   SwapBuffers(g_HDC); // Display the new frame.
}


void GetInput()
{
   // Update the time for time based movement.
   Camera.CalculateTime();

   // Here we check for each input that we care about.
   if(GetKeyState(VK_UP) & 0x80) Camera.MoveCamera(UP);
   if(GetKeyState(VK_DOWN) & 0x80) Camera.MoveCamera(DOWN);

   if(GetKeyState(VK_LEFT) & 0x80)
      Camera.RotateCamera(CVector4(Camera.View.x, Camera.View.y, Camera.View.z),
                          LEFT, 0, 1, 0);          // Rotate the camera left in 3rd person.

   if(GetKeyState(VK_RIGHT) & 0x80)
      Camera.RotateCamera(CVector4(Camera.View.x, Camera.View.y, Camera.View.z),
                          RIGHT, 0, 1, 0);         // Rotate the camera right in 3rd person.

   if(GetKeyState('W') & 0x80 || GetKeyState('w') & 0x80)
      Camera.RotateCamera(CVector4(Camera.View.x, Camera.View.y, Camera.View.z),
                          UP, 1, 0, 0);            // Rotate the camera up in 3rd person.

   if(GetKeyState('S') & 0x80 || GetKeyState('s') & 0x80)
      Camera.RotateCamera(CVector4(Camera.View.x, Camera.View.y, Camera.View.z),
                          DOWN, 1, 0, 0);          // Rotate the camera down in 3rd person.
}


void Shutdown()
{
   // Free both textures that we loaded earlier.
   Terrain.FreeImage();
   CubeImage.FreeImage();
   glDeleteTextures(1, &Terrain.ID);
   glDeleteTextures(1, &CubeImage.ID);
}


// Recap:

// Now we have a third person view as well as a first person view for our camera class.
// This camera class can use a lot more bulk but as it is it covers the basics pretty well.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com